@@ -4,6 +4,8 @@ from .grid import (
     enable,
     disable,
     set_defaults,
+    on,
+    off,
     set_grid_option,
     show_grid,
     QgridWidget,
@@ -22,6 +24,8 @@ __all__ = [
     'enable',
     'disable',
     'set_defaults',
+    'on',
+    'off',
     'set_grid_option',
     'show_grid',
     'QgridWidget',
